/* xshostakovich.cc
   Copyright (C) 1998-9 Eddie Kohler, eddietwo@lcs.mit.edu
   This file is part of PORTABLE XSHOSTAKOVICH++ 98.

   PORTABLE XSHOSTAKOVICH++ 98 is free software. It is distributed under the
   GNU Public License, version 2 or later; you can copy, distribute, or alter
   it at will, as long as this notice is kept intact, this source code is made
   available, and you are crazy nuts. Warranty? Ha! */

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif
#include "xshostakovich.hh"
#include "eyesmask.xbm"
#include "clp.h"
#include <string.h>
#include <stdio.h>
#include <ctype.h>
#include <X11/Xutil.h>
#include <X11/Xresource.h>
#include <X11/keysym.h>

// Image records (in images.c)
extern Gif_Record shostakovich_gif, smile_gif, tongue_gif, lefteye_gif,
  righteye_gif, lefteyeball_gif, righteyeball_gif;

static Atom wm_protocols_atom;
static Atom wm_delete_window_atom;


GifPicture::GifPicture(Gif_Record *gifrec, Gif_XContext *gfx)
  : _gfx(gfx), _pixmap(None)
{
  _gfs = Gif_FullReadRecord
    (gifrec, GIF_READ_COMPRESSED | GIF_READ_CONST_RECORD, 0, 0);
}

void
GifPicture::create()
{
  _pixmap = Gif_XImage(_gfx, _gfs, 0);
}


const char *Shostakovich::saying = 0;
const char *Shostakovich::font_name = 0;
const char *Shostakovich::saying_color_name = 0;
bool Shostakovich::quiet = false;

Shostakovich::Shostakovich(Display *display, Window window, Colormap colormap)
  : _display(display), _window(window), _colormap(colormap),
    _gfx(Gif_NewXContext(display, window)),
    _tongueing(false), _left_eye_dead(false), _right_eye_dead(false),
    _just_iconified(0)
{
  _dsch = new GifPicture(&shostakovich_gif, _gfx);
  _smile = new GifPicture(&smile_gif, _gfx);
  _tongue = new GifPicture(&tongue_gif, _gfx);
  _left_eye = new GifPicture(&lefteye_gif, _gfx);
  _right_eye = new GifPicture(&righteye_gif, _gfx);
  _left_eyeball = new GifPicture(&lefteyeball_gif, _gfx);
  _right_eyeball = new GifPicture(&righteyeball_gif, _gfx);

  {
    Window root;
    int x, y;
    unsigned border_width, depth;
    XGetGeometry(_display, _window, &root, &x, &y, &_width, &_height,
		 &border_width, &depth);
  }
  _origin_x = (int)(_width - _dsch->width()) / 2;
  _origin_y = (int)(_height - _dsch->height()) / 2;
  
  XColor crap_color;
  
  // get the window's background color
  if (XAllocColorCells(_display, _colormap, True, 0, 0,
		       &_background.pixel, 1) != 0) {
    _background_mutable = true;
    //unsigned long pixel = _background.pixel;
    XParseColor(_display, _colormap, BackgroundColor, &_background);
    XStoreColor(_display, _colormap, &_background);
  } else if (XAllocNamedColor(_display, _colormap, BackgroundColor,
			      &_background, &crap_color) != 0) {
    _background_mutable = false;
  } else {
    _background_mutable = false;
    _background.pixel = BlackPixel(_display, DefaultScreen(_display));
  }
  XSetWindowBackground(_display, _window, _background.pixel);
  
  // Set up the GC
  XGCValues gcv;
  unsigned long gcv_mask = 0;
  
  XColor xcolor;
  if (XAllocNamedColor(_display, _colormap, saying_color_name, &xcolor,
		       &crap_color) != 0) {
    gcv.foreground = xcolor.pixel;
    gcv_mask |= GCForeground;
  }
  
  XFontStruct *font = 0;
  if (font_name != 0)
    font = XLoadQueryFont(_display, font_name);
  if (font == 0)
    font = XLoadQueryFont(_display, "-adobe-itc souvenir-demi-i-*-*-*-240-*-*-*-*-*-*");
  if (font == 0)
    font = XLoadQueryFont(_display, "-monotype-bembo-bold-i-*-*-*-240-*-*-*-*-*-*");
  if (font == 0)
    font = XLoadQueryFont(_display, "-adobe-utopia-bold-i-*-*-*-240-*-*-*-*-*-*");
  if (font == 0)
    font = XLoadQueryFont(_display, "fixed");
  gcv.font = font->fid;
  gcv_mask |= GCFont;
  
  _gc = XCreateGC(_display, _window, gcv_mask, &gcv);
  
  _saying = saying;
  _saying_width = XTextWidth(font, _saying, strlen(_saying));
  _quiet = quiet;
  
  // get rid of the XFontStruct
  XFreeFont(_display, font);
  
  // The eye-mask for drawing the eyeballs. We set the clip_mask portion
  // of the gcontext used for drawing the eyeballs (when they follow the
  // mouse around) to a bitmap (created previously) which is black
  // everywhere, except for in Shostakovich's eyeballs! That way, when we
  // draw the eyeballs, the only pixels which will be affected will be
  // the pixels in Shostakovich's eyeballs -- which is just right.
  Pixmap clip_mask = XCreateBitmapFromData(_display, _window, eyesmask_bits,
					   eyesmask_width, eyesmask_height);
  gcv.clip_x_origin = _origin_x + EyesMaskOffsetX;
  gcv.clip_y_origin = _origin_y + EyesMaskOffsetY;
  gcv.clip_mask = clip_mask;
  _eye_gc = XCreateGC(_display, _window, GCClipXOrigin | GCClipYOrigin
		      | GCClipMask, &gcv);

  XDefineCursor(_display, _window, XCreateFontCursor(_display, 88));
  XClearWindow(_display, _window);
}


void
Shostakovich::draw()
{
  // effects  Draws Shostakovich (the main pixmap) and the saying onto
  //   s.mainw.
  
  // Note: we offset everything by the s.orig? values, so that we can
  // keep Shostakovich centered when the window's size changes.
  unsigned right = _origin_x + _dsch->width();
  unsigned bottom = _origin_y + _dsch->height();

  if (_height > _dsch->height()) {
    XClearArea(_display, _window, 0, 0, _width, _origin_y, False);
    XClearArea(_display, _window, 0, bottom, _width, _height - bottom, False);
  }
  if (_width > _dsch->width()) {
    XClearArea(_display, _window, 0, _origin_y,
	       _origin_x, _dsch->height(), False);
    XClearArea(_display, _window, right, _origin_y,
	       _width - right, _dsch->height(), False);
  }
  
  XCopyArea(_display, _dsch->pixmap(), _window, _gc,
	    0, 0, _dsch->width(), _dsch->height(),
	    _origin_x, _origin_y);
  
  if (_left_eye_dead) draw_left_eye(true);
  if (_right_eye_dead) draw_right_eye(true);
  
  draw_string(216 - _saying_width, 372, _saying);
}

void
Shostakovich::draw_part(GifPicture *part, bool draw, int x_off, int y_off)
{
  if (draw)
    XCopyArea(_display, part->pixmap(), _window, _gc,
	      0, 0, part->width(), part->height(),
	      _origin_x + x_off, _origin_y + y_off);
  else
    XCopyArea(_display, _dsch->pixmap(), _window, _gc,
	      x_off, y_off, part->width(), part->height(),
	      _origin_x + x_off, _origin_y + y_off);
}


void
Shostakovich::draw_smile(bool smiling)
{
  if (_tongueing)
    draw_tongue(false);
  draw_part(_smile, smiling, SmileOffsetX, SmileOffsetY);
}

void
Shostakovich::draw_tongue(bool tongueing)
{
  // modifies  s.
  // effects  Draws Shostakovich's tongue sticking out if tongueing is
  //   true, or redraws his normal, blank expression if tongueing is 
  //   false. Sets the s.tongueing field to the value of tongueing.
  draw_part(_tongue, tongueing, TongueOffsetX, TongueOffsetY);
  _tongueing = tongueing;
}

void
Shostakovich::draw_left_eye(bool xed_out)
{
  // modifies  s.
  // effects  X'es out Shostakovich's left eye if xed_out is true, or redraws
  //   it as normal if xed_out is false. Sets s.lefteyedead to out.
  draw_part(_left_eye, xed_out, LeftEyeOffsetX, LeftEyeOffsetY);
  _left_eye_dead = xed_out;
}

void
Shostakovich::draw_right_eye(bool xed_out)
{
  // modifies  s.
  // effects  X'es out Shostakovich's right eye if out is true, or redraws
  //   it as normal if out is false. Sets s.righteyedead to out.
  draw_part(_right_eye, xed_out, RightEyeOffsetX, RightEyeOffsetY);
  _right_eye_dead = xed_out;
}


void
Shostakovich::follow_right_eye(const XMotionEvent &motev)
{
  // effects  Makes Shostakovich's right eye appear to be looking at the
  //   mouse pointer. The mouse pointer's position is specified by the
  //   motion_event motev.
	
  if (_right_eye_dead) return;
  
  // Lots of calculations to make it work semi-realistically...
  // (The numbers were found by trial and error.)
  int widbearing = RightEyeWidth  / 2;
  int hgtbearing = RightEyeHeight / 2;
  int deltax = (RightEyeOffsetX + widbearing - motev.x + _origin_x) / 20;
  int deltay = (RightEyeOffsetY + hgtbearing - motev.y + _origin_y) / 20;
  if (deltax < -widbearing+25) deltax = -widbearing+25;
  if (deltax >  widbearing-24) deltax =  widbearing-24;
  if (deltay < -hgtbearing+30) deltay = -hgtbearing+30;
  if (deltay >  hgtbearing-30) deltay =  hgtbearing-30;

  XCopyArea(_display, _right_eyeball->pixmap(), _window, _eye_gc,
	    0, 0, _right_eyeball->width(), _right_eyeball->height(),
	    _origin_x + RightEyeballOffsetX - deltax,
	    _origin_y + RightEyeballOffsetY - deltay);
}
	
	
void
Shostakovich::follow_left_eye(const XMotionEvent &motev)
{
  // effects  Makes Shostakovich's left eye appear to be looking at the
  //   mouse pointer. The mouse pointer's position is specified by the
  //   motion_event motev.
	
  if (_left_eye_dead) return;
  
  // Lots of calculations to make it work semi-realistically...
  // (The numbers were found by trial and error.)
  int widbearing = LeftEyeWidth  / 2;
  int hgtbearing = LeftEyeHeight / 2;
  int deltax = (LeftEyeOffsetX + widbearing - motev.x + _origin_x) / 20;
  int deltay = (LeftEyeOffsetY + hgtbearing - motev.y + _origin_y) / 20;
  if (deltax < -widbearing+15) deltax = -widbearing+15;
  if (deltax >  widbearing-20) deltax =  widbearing-20;
  if (deltay < -hgtbearing+20) deltay = -hgtbearing+20;
  if (deltay >  hgtbearing-24) deltay =  hgtbearing-24;

  XCopyArea(_display, _left_eyeball->pixmap(), _window, _eye_gc,
	    0, 0, _left_eyeball->width(), _left_eyeball->height(),
	    _origin_x + LeftEyeballOffsetX - deltax,
	    _origin_y + LeftEyeballOffsetY - deltay);
}


void
Shostakovich::draw_string(int dx, int dy, const char *str)
{
  if (!_quiet)
    XDrawString(_display, _window, _gc, dx + _origin_x, dy + _origin_y,
		str, strlen(str));
}

void
Shostakovich::do_drag()
{
  // effects  Implements the useless drag operation. The pointer is
  //   grabbed, the cursor is changed to a spider, and Shostakovich's
  //   eyes follow the pointer around the screen no matter where it goes.
  //   The grab is released, the pointer returns to normal, etc. when the
  //   user releases mouse button 2.
  
  XGrabPointer(_display, _window, False,
	       ButtonReleaseMask | PointerMotionMask,
	       GrabModeAsync, GrabModeAsync,
	       None, XCreateFontCursor(_display, 122), CurrentTime);
  
  while (true) {
    XEvent e;
    XNextEvent(_display, &e);
    if (e.type == ButtonRelease && e.xbutton.button == 2)
      break;
    else if (e.type == MotionNotify) {
      follow_right_eye(e.xmotion);
      follow_left_eye(e.xmotion);
    }
  }
  
  XUngrabPointer(_display, CurrentTime);
}


bool
Shostakovich::in_dsch_rect(int dx, int dy, int x, int y, int w, int h) const
{
  dx -= _origin_x;
  dy -= _origin_y;
  return (dx >= x && dy >= y && dx < x + w && dy < y + h);
}


void
Shostakovich::die()
{
  int endy = _origin_y + _dsch->height();
  for (int y = _origin_y; y < endy; y += 10) {
    XCopyArea(_display, _window, _window, _gc,
	      _origin_x, y, _dsch->width(), endy - y - 10,
	      _origin_x, y + 10);
    XSync(_display, False);
  }
}


bool
Shostakovich::handle_event(XEvent &e)
{
  if (e.type != MappingNotify && e.xany.window != _window)
    return true;
  
  switch (e.type) {
    
   case Expose:
    if (e.xexpose.count == 0) {
      draw();
      if (_just_iconified) {
	// If we were mapped after an attempted iconification,
	// put a little snide message up in addition to everything else.
	draw_string(10, 30, "Ha! Ha! Fool!");
	_just_iconified--;
      }
    }
    break;

   case KeyPress:
     {
       KeySym keysym;
       char buffer[5];
       XComposeStatus compose;
       int nbuffer = XLookupString(&e.xkey, buffer, 5, &keysym, &compose);
       
       if (keysym == XK_Delete || keysym == XK_BackSpace)
	 draw_string(15, 80, "You can't delete me!");
       if (nbuffer != 1)
	 break;
       
       if (buffer[0] == 'w') {
	 XWarpPointer(_display, None, None, 0, 0, 0, 0, -2, 2);
       } else if (buffer[0] == 'W') {
	 XWarpPointer(_display, None, None, 0, 0, 0, 0, 2, -2);
       } else if (buffer[0] == 'c') {
	 if (_background_mutable) {
	   _background.red += 300;
	   _background.green -= 300;
	   _background.blue += 200;
	   XStoreColor(_display, _colormap, &_background);
	 } else
	   XBell(_display, 0);
       }
       
       break;
     }

   case EnterNotify:
    draw_smile(true);
    break;

   case LeaveNotify:
    draw_smile(false);
    break;

   case ButtonPress:
    draw_tongue(true);
    if (e.xbutton.button == 2)
      do_drag();
    break;

   case MotionNotify:
    follow_right_eye(e.xmotion);
    follow_left_eye(e.xmotion);
    break;
    
   case ButtonRelease:
    if (in_dsch_rect(e.xbutton.x, e.xbutton.y, LeftEyeOffsetX, LeftEyeOffsetY,
		     LeftEyeWidth, LeftEyeHeight)) {
      // Check for knocking out the left eye...
      draw_left_eye(!_left_eye_dead);
      draw_smile(!_left_eye_dead);
    } else if (in_dsch_rect(e.xbutton.x, e.xbutton.y, RightEyeOffsetX,
			    RightEyeOffsetY, RightEyeWidth, RightEyeHeight)) {
      // Check for knocking out the right eye...
      draw_right_eye(!_right_eye_dead);
      draw_smile(!_right_eye_dead);
    } else
      draw_tongue(false);
    if (_left_eye_dead && _right_eye_dead) {
      // If both eyes are now dead, let's quit!
      draw_tongue(true);
      die();
      return false;
    }
    break;
    
   case UnmapNotify:
    // We are unmapped mostly when the user tries to iconify us.
    // We are contrary and don't like to be iconified. Ha!
    XMapRaised(_display, _window);
    _just_iconified = 2;
    break;
    
   case ConfigureNotify:
    // Make sure Dmitri appears in the center of the window
    _width = e.xconfigure.width;
    _height = e.xconfigure.height;
    _origin_x = (int)(_width - _dsch->width()) / 2;
    _origin_y = (int)(_height - _dsch->height()) / 2;
    XSetClipOrigin(_display, _eye_gc, _origin_x + EyesMaskOffsetX,
		   _origin_y + EyesMaskOffsetY);
    draw();
    break;
    
   case ClientMessage:
    if (e.xclient.message_type == wm_protocols_atom &&
	(Atom)(e.xclient.data.l[0]) == wm_delete_window_atom)
      draw_string(30, 80, "You can't delete me!");
    break;

   case MappingNotify:
    XRefreshKeyboardMapping(&e.xmapping);
    break;
    
  }
  
  return true;
}


/*****
 * Helper functions and main program
 **/

static int on_root = -1;
static int infinity = -1;
static bool set_quiet = false;
static char *geometry_spec = 0;
static const char *program_name;

static bool
unparse_bool(const char *in_str)
{
  if (strlen(in_str) > 5) return false;
  char buf[6], *s;
  for (s = buf; *in_str; in_str++, s++)
    *s = tolower(*in_str);
  *s = 0;
  return (strcmp(buf, "true") == 0 || strcmp(buf, "1") == 0
	  || strcmp(buf, "yes") == 0 || strcmp(buf, "y") == 0);
}

static void
handle_resources(XrmDatabase xrmdb)
{
  char *str_type;
  XrmValue value;
  
  if (on_root == -1) {
    if (XrmGetResource(xrmdb, "xshostakovich.root", "XShostakovich.Root",
		       &str_type, &value))
      on_root = (unparse_bool(value.addr) ? 1 : 0);
    else
      on_root = 0;
  }
  
  if (!geometry_spec) {
    if (XrmGetResource(xrmdb, "xshostakovich.geometry",
		       "XShostakovich.Geometry", &str_type, &value))
      geometry_spec = value.addr;
  }
  
  if (infinity == -1) {
    if (XrmGetResource(xrmdb, "xshostakovich.infinity",
		       "XShostakovich.Infinity", &str_type, &value))
      infinity = (unparse_bool(value.addr) ? 1 : 0);
    else
      infinity = 0;
  }
  
  if (!Shostakovich::saying) {
    if (XrmGetResource(xrmdb, "xshostakovich.saying", "XShostakovich.Saying",
		       &str_type, &value))
      Shostakovich::saying = value.addr;
    else
      Shostakovich::saying = "Hello, Dmitri";
  }
  
  if (!set_quiet)
    if (XrmGetResource(xrmdb, "xshostakovich.quiet", "XShostakovich.Quiet",
		       &str_type, &value))
      Shostakovich::quiet = unparse_bool(value.addr);
  
  if (!Shostakovich::saying_color_name) {
    if (XrmGetResource(xrmdb, "xshostakovich.textColor",
		       "XShostakovich.TextColor", &str_type, &value) == True)
      Shostakovich::saying_color_name = value.addr;
    else
      Shostakovich::saying_color_name = "MediumSeaGreen";
  }
  
  if (!Shostakovich::font_name)
    if (XrmGetResource(xrmdb, "xshostakovich.font", "XShostakovich.Font",
		       &str_type, &value))
      Shostakovich::font_name = value.addr;
}


static int
parse_geometry(const char *const_g, XSizeHints *sh, int screen_width,
	       int screen_height)
{
  char *g = (char *)const_g;
  sh->flags = 0;
  
  if (isdigit(*g)) {
    sh->flags |= USSize;
    sh->width = strtol(g, &g, 10);
    if (g[0] == 'x' && isdigit(g[1]))
      sh->height = strtol(g + 1, &g, 10);
    else
      goto error;
  } else if (!*g)
    goto error;
  
  if (*g == '+' || *g == '-') {
    int x_minus, y_minus;
    sh->flags |= USPosition | PWinGravity;
    x_minus = *g == '-';
    sh->x = strtol(g + 1, &g, 10);
    if (x_minus) sh->x = screen_width - sh->x - sh->width;
    
    y_minus = *g == '-';
    if (*g == '-' || *g == '+')
      sh->y = strtol(g + 1, &g, 10);
    else
      goto error;
    if (y_minus) sh->y = screen_height - sh->y - sh->height;
    
    if (x_minus)
      sh->win_gravity = y_minus ? SouthEastGravity : NorthEastGravity;
    else
      sh->win_gravity = y_minus ? SouthWestGravity : NorthWestGravity;
    
  } else if (*g)
    goto error;

  return 1;
  
 error:
  fprintf(stderr, "%s: bad geometry specification\n", program_name);
  sh->flags = 0;
  return 0;
}

static Window
create_top_level(Display *display, int screen, Colormap colormap)
{
  XClassHint classh;
  XTextProperty window_name, icon_name;
  XSetWindowAttributes setattr;
  unsigned long setattr_mask;
  
  // Set up window name and icon name
  {
    char *string_list[2];
    string_list[0] = "xshostakovich";
    string_list[1] = 0;
    XStringListToTextProperty(string_list, 1, &window_name);
    XStringListToTextProperty(string_list, 1, &icon_name);
    classh.res_name = "xshostakovich";
    classh.res_class = "XShostakovich";
  }
  
  setattr.colormap = colormap = DefaultColormap(display, screen);
  setattr.backing_store = NotUseful;
  setattr.save_under = False;
  setattr.border_pixel = 0;
  setattr.background_pixel = 0;
  setattr_mask = CWColormap | CWBorderPixel | CWBackPixel | CWBackingStore
    | CWSaveUnder;
  
  XSizeHints *xsizeh = XAllocSizeHints();
  xsizeh->width = 640;
  xsizeh->height = 520;
  if (geometry_spec)
    parse_geometry(geometry_spec, xsizeh, DisplayWidth(display, screen),
		   DisplayHeight(display, screen));
  
  Window window = XCreateWindow
    (display, DefaultRootWindow(display),
     xsizeh->x, xsizeh->y, xsizeh->width, xsizeh->height, 0,
     DefaultDepth(display, screen), InputOutput,
     DefaultVisual(display, screen),
     setattr_mask, &setattr);
  
  XSetWMProperties(display, window, &window_name, &icon_name,
		   NULL, 0, xsizeh, 0, &classh);
  XSetWMProtocols(display, window, &wm_delete_window_atom, 1);
  
  XSelectInput(display, window, ExposureMask | EnterWindowMask
	       | LeaveWindowMask | KeyPressMask | KeyReleaseMask
	       | ButtonPressMask | ButtonReleaseMask | PointerMotionMask
	       | StructureNotifyMask);
  
  XMapRaised(display, window);
  
  XFree(xsizeh);
  XFree(window_name.value);
  XFree(icon_name.value);
  return window;
}


enum {
  XRM_OPT = 300, FONT_OPT, ROOT_OPT, TEXT_COLOR_OPT, DISPLAY_OPT, SAYING_OPT,
  QUIET_OPT, VERSION_OPT, HELP_OPT, INFINITY_OPT, GEOMETRY_OPT
};

static Clp_Option options[] = {
  { "display", 'd', DISPLAY_OPT, Clp_ArgStringNotOption, 0 },
  { "fn", 'f', FONT_OPT, Clp_ArgString, 0 },
  { "font", 0, FONT_OPT, Clp_ArgString, 0 },
  { "geometry", 0, GEOMETRY_OPT, Clp_ArgString, 0 },
  { "help", 'h', HELP_OPT, 0, 0 },
  { "infinity", 'i', INFINITY_OPT, 0, Clp_Negate },
  { "quiet", 'q', QUIET_OPT, 0, 0 },
  { "root", 'r', ROOT_OPT, 0, Clp_Negate },
  { "saying", 's', SAYING_OPT, Clp_ArgString, Clp_Negate },
  { "text-color", 'c', TEXT_COLOR_OPT, Clp_ArgString, 0 },
  { "version", 'v', VERSION_OPT, 0, 0 },
  { "xrm", 0, XRM_OPT, Clp_ArgString, 0 },
};

static void
short_usage()
{
  fprintf(stderr, "Usage: %s [OPTION]...\n\
Try `%s --help' for more information.\n",
	  program_name, program_name);
}

static void
usage()
{
  printf("\
`Portable Xshostakovich++ 98' is a stupid X display trick. Xshostakovich was\n\
originally written to test a CLU/Xlib interface. (Yeah!!)\n\
\n\
Usage: %s [OPTION]...\n\
\n\
Options are:\n\
  -display DISPLAY              Tell Dmitri S. where to live\n\
  -geometry GEOMETRY            Put his window somewhere special\n\
  -root, -r | -no-root          Display him on the root window, or not\n\
  -fn FONT                      Change his handwriting to FONT\n\
  -text-color COLOR             Give him COLORed ink\n\
  -saying PITHY-SAYING          Dmitri S. will pithily say PITHY-SAYING\n\
  -quiet, -q | -no-quiet        He will keep his own counsel, or not\n\
  -infinity | -no-infinity      Shostakovich forever! (or not)\n\
  -xrm RESOURCE                 Tell him how to behave\n\
  --help, -h                    Print this message and exit\n\
  --version, -v                 Print version number and exit\n\
\n\
Reporting bugs? <eddietwo@lcs.mit.edu>\n\
Don't like my driving? www.eat-shit.com\n", program_name);
}


int
main(int argc, char **argv)
{
  char *display_name = 0;
  char **saved_resources = new char *[argc];
  int nsaved_resources = 0;
  
  // handle command line
  Clp_Parser *clp =
    Clp_NewParser(argc, argv, sizeof(options) / sizeof(Clp_Option), options);
  Clp_SetOptionChar(clp, '-', Clp_Long | Clp_Short);
  Clp_SetOptionChar(clp, '+', Clp_LongNegated | Clp_ShortNegated);
  
  program_name = Clp_ProgramName(clp);
  // special handling for `xshostakovichinfinity'
  if (strcmp(program_name, "xshostakovichinfinity") == 0)
    infinity = 1;
  
  while (1) {
    int opt = Clp_Next(clp);
    switch (opt) {
      
     case XRM_OPT:
      saved_resources[nsaved_resources] = clp->arg;
      nsaved_resources++;
      break;
      
     case DISPLAY_OPT:
      display_name = clp->arg;
      break;

     case GEOMETRY_OPT:
      geometry_spec = clp->arg;
      break;
      
     case SAYING_OPT:
      Shostakovich::saying = (clp->negated ? "" : clp->arg);
      break;
      
     case FONT_OPT:
      Shostakovich::font_name = clp->arg;
      break;
      
     case TEXT_COLOR_OPT:
      Shostakovich::saying_color_name = clp->arg;
      break;
      
     case ROOT_OPT:
      on_root = (clp->negated ? 0 : 1);
      break;

     case INFINITY_OPT:
      infinity = (clp->negated ? 0 : 1);
      break;
      
     case QUIET_OPT:
      Shostakovich::quiet = (clp->negated ? false : true);
      set_quiet = true;
      break;
      
     case VERSION_OPT:
      printf("PORTABLE XSHOSTAKOVICH++ 98 version %s\n", VERSION);
      printf("Copyright (C) 1992-2000 Eddie Kohler\n\
This is free software; see the source for copying conditions.\n\
A warranty in the secret language of dance is available from the publishers.\n\
To summarize its contents: If this product does not fit your requirements,\n\
including, but not limited to, any requirements for merchantability or\n\
fitness for a particular purpose, you are entitled to a melancholy feeling.\n");
      exit(0);
      break;
      
     case HELP_OPT:
      usage();
      exit(0);
      break;
      
     case Clp_Done:
      goto done;
      
     case Clp_NotOption:
      fprintf(stderr, "%s: I don't understand `%s'.\n", program_name,
	      clp->arg);
      /* FALLTHRU */
      
     case Clp_BadOption:
      short_usage();
      exit(1);
      break;
      
    }
  }

 done:
  Display *display = XOpenDisplay(display_name);
  int screen = DefaultScreen(display);
  Colormap colormap = DefaultColormap(display, screen);
  
  wm_protocols_atom = XInternAtom(display, "WM_PROTOCOLS", False);
  wm_delete_window_atom = XInternAtom(display, "WM_DELETE_WINDOW", False);
  
  XrmInitialize();
  XrmDatabase xrmdb = 0;
  if (XResourceManagerString(display))
    xrmdb = XrmGetStringDatabase(XResourceManagerString(display));
  if (!xrmdb)
    xrmdb = XrmGetStringDatabase("");
  for (int i = 0; i < nsaved_resources; i++)
    XrmPutLineResource(&xrmdb, saved_resources[i]);

  handle_resources(xrmdb);

  do {
    Window window;
    if (on_root) {
      window = RootWindow(display, screen);
      XSelectInput(display, window, ExposureMask | EnterWindowMask
		   | LeaveWindowMask | PointerMotionMask);
    } else
      window = create_top_level(display, screen, colormap);
    
    Shostakovich shostakovich(display, window, colormap);
    if (on_root) shostakovich.draw();
    XEvent event;
    while (1) {
      XNextEvent(display, &event);
      if (!shostakovich.handle_event(event)) break;
    }
    
    if (!on_root)
      XDestroyWindow(display, window);
  } while (infinity && !on_root);
  
  return 0;
}
